//
//  Test_v3cross.h
//  BulletTest
//
//  Copyright (c) 2011 Apple Inc.
//

#ifndef BulletTest_Test_v3cross_h
#define BulletTest_Test_v3cross_h

#ifdef __cplusplus
extern "C" { 
#endif

int Test_v3cross(void);

#ifdef __cplusplus
}
#endif

    
#endif
